#!/usr/bin/env python

__version__ = '$Revision: 5698 $'.split()[1]
__date__ = '$Date: 2007-03-03 12:58:28 -0500 (Sat, 03 Mar 2007) $'.split()[1]
__author__ = 'Kurt Schwehr'

__doc__='''
Handle creation and extraction of NMEA strings.  Maybe need a separate VDM class like ais-py?

@requires: U{lxml<http://codespeak.net/lxml/>} - For libxml2 ElementTree interface.  Not actually required for the template, but this is just a demo or requirements.
@requires: U{Python<http://python.org/>} >= 2.4
@requires: U{epydoc<http://epydoc.sourceforge.net/>} >= 3.0alpha3
@requires: BitVector

@author: U{'''+__author__+'''<http://schwehr.org/>}
@version: ''' + __version__ +'''
@copyright: 2006
@var __date__: Date of last svn commit
@undocumented: __version__ __author__ __doc__ myparser
@since: 2006-Sep-26  FIX: replace with the file creation date
@status: under development
@organization: U{CCOM<http://ccom.unh.edu/>} - FIX: if not CCOM change the name and link

@license: GPL v2

@note: This package does not respence the maximum number of characters
per line that is required in the NMEA specification.

'''

# Python standard libraries
import time, sys

# Local Modules
import binary

import re

EOL = "\x0D\x0A"
'''
DOS style end-of-line (<cr><lf>) for talking to AIS base stations
'''


def checksumStr(data,verbose=False):
    """
    Take a NMEA 0183 string and compute the checksum.

    Checksum is calculated by xor'ing everything between ? or ! and the *

    >>> checksumStr("!AIVDM,1,1,,B,35MsUdPOh8JwI:0HUwquiIFH21>i,0*09")
    '09'

    >>> checksumStr("AIVDM,1,1,,B,35MsUdPOh8JwI:0HUwquiIFH21>i,0")
    '09'

    >>> checksumStr('$AIACA,0,,,,,,,,,5,2087,0,2088,0,0,0,I,1,000000*15')
    '15'

    This is an example I made up

    >>> checksumStr('$xxCAB,1,1,1,1*5D')
    '40'

    @param data: NMEA message.  Leading ?/! and training checksum are optional
    @type data: str
    @return: hexidecimal value
    @rtype: str

    """

    # FIX: strip off new line at the end too
    #if data[0]=='!' or data[0]=='?': data = data[1:]
    #if data[-1]=='*': data = data[:-1]
    #if data[-3]=='*': data = data[:-3]
    end = data.find('*') # FIX: would rfind be faster?
    start=0
    if data[0] in ('$','!'): start=1
    if -1 != end: data=data[start:end]
    else: data=data[start:]
    if verbose: print 'checking on:',start,end,data
    # FIX: rename sum to not shadown builting function
    sum=0
    for c in data: sum = sum ^ ord(c)
    sumHex = "%x" % sum
    if len(sumHex)==1: sumHex = '0'+sumHex
    return sumHex.upper()


######################################################################
# common variables
#nmeaChecksumRegExStr = r"""\,[0-9]\*[0-9A-F][0-9A-F]"""
#nmeaChecksumRegExStr = r"""\,[A-Za-z0-9]\*[0-9A-F][0-9A-F]"""
nmeaChecksumRegExStr = r"""\*[0-9A-F][0-9A-F]"""
nmeaChecksumRE = re.compile(nmeaChecksumRegExStr)

def isChecksumValid(nmeaStr, allowTailData=True,verbose=False):
    """Return True if the string checks out with the checksum


    >>> isChecksumValid("!AIVDM,1,1,,B,35MsUdPOh8JwI:0HUwquiIFH21>i,0*09")
    True

    Corrupted:

    >>> isChecksumValid("!AIVDM,11,1,,B,35MsUdPOh8JwI:0HUwquiIFH21>i,0*09")
    False

    >>> isChecksumValid('$AIACA,0,,,,,,,,,5,2087,0,2088,0,0,0,I,1,000000*15')
    True

    @param allowTailData: Permit handing of Coast Guard format with data after the checksum
    @param nmeaStr: NMEA message.  Leading ?/! are optional
    @type nmeaStr: str
    @return: True if the checksum matches
    @rtype: bool
    """

    if allowTailData:
	match = nmeaChecksumRE.search(nmeaStr)
	if not match: 
            if verbose: print 'Match failed'
            return False
	nmeaStr = nmeaStr[:match.end()]
	#if checksum.upper()==checksumStr(nmeaStr[match.end()


    if nmeaStr[-3]!='*': 
	print 'FIX: warning... bad nmea string'
	return False  # Bad string without proper checksum
    checksum=nmeaStr[-2:]
    if checksum.upper()==checksumStr(nmeaStr).upper():
        return True
    if verbose: 
        print 'mismatch checksums:', checksum.upper(),checksumStr(nmeaStr).upper()
    return False

def buildNmea(aisBits,prefix='!',serviceType='AI',msgType='VDM',channelSeq=None,channel='A'):
    '''
    Create one long oversized nmea string for the bits
    @param aisBits: message payload
    @type aisBits: BitVector
    @param prefix: '!' or '$'  what is the difference?
    @param serviceType: 'can this be anything other than AI?
    @param msgType: VDM.  Should not be VDO (own ship)
    @param channelSeq: 1-9 or None
    @param channel: AIS channel A or B
    @todo: sync names of prefix and serviceType to NMEA spec.
    @see: reference the appropriate spec documents for all this stuff.
    '''

    rList = [prefix,serviceType,msgType,',1,1,']
    if None != channelSeq: rList.append(str(channelSeq))
    rList.append(',')
    rList.append(channel)
    rList.append(',')

    payloadStr,pad = binary.bitvectoais6(aisBits) #[0]
    rList.append(payloadStr)
    rList.append(','+str(pad))
    rStr = ''.join(rList)
    rStr += '*'+checksumStr(rStr)

    return rStr


######################################################################
def cabEncode(TransA=False, TransB=False, Restart=False, Reset=False, prefix='AI'):  # FIX: default to xx?
    '''
    CAB - Control AIS Base Station.  Defaults to a safe state with
    everything shutdown.  62320-1/CDV, 80/427/CDV, Page 77, A.1.7

    >>> cabEncode()
    '$AICAB,0,0,,*48'

    Note that xx is probably not valid in this next example, but it is used by L3

    Made up example:

    >>> cabEncode(True,True,True,True,prefix='xx')
    '$xxCAB,1,1,1,1*40'

    @param TransA: Transmissions enabled on channel A
    @type TransA: bool
    @param TransB: Transmissions enabled on channel B
    @type TransB: bool
    @param Restart: If true, command AIS Base station to restart operations in last known configuration
    @type Restart: bool
    @param Reset: 
    @type Reset: bool
    @param prefix: string to put between the $ and CAB
    @type prefix: str
    @return: A CAB NMEA string
    @rtype: str
    '''
    r = ['$'+prefix+'CAB']
    if TransA: r.append('1')
    else: r.append('0')
    if TransB: r.append('1')
    else: r.append('0')
    if Restart: r.append('1')
    else: r.append('')
    if Reset: r.append('1')
    else: r.append('')
    rStr = ','.join(r)
    return rStr+'*'+checksumStr(rStr)


def cabDecode(msg,validate=True):
    '''

    >>> cabDecode('$AICAB,,,,*48')
    {'Reset': False, 'nmeaPrefix': 'AI', 'nmeaCmd': 'CAB', 'TransB': False, 'TransA': False, 'Restart': False}

    Note that ZZ is probably not valid in this next example

    >>> cabDecode('$ZZCAB,1,1,1,1*40')
    {'Reset': False, 'nmeaPrefix': 'ZZ', 'nmeaCmd': 'CAB', 'TransB': True, 'TransA': True, 'Restart': True}

    @param msg: NMEA string of a CAB message
    @type msg: str
    @param validate: Set to False to turn off validation for speed.
    @type validate: bool
    @return: lookup table of key/values
    @rtype: dict

    @todo: How do I make stable doctests with dictionary returns
    @todo FIX: throw an exception if not valid
    '''
    if validate and not isChecksumValid(msg,verbose=True):
        print 'FIX: this should be an exception in cabDecode.  Bad checksum'
        return False
    fields=msg.split(',')
    if validate and len(fields) not in (5,6,7): # Allow for USCG station and timestamp
        # check for USCG log tail
        print 'FIX: this should be an exception in cabDecode.  wrong number of fields'
        return False
        
    # FIX: for validate... make sure that the other case from 1 is an empty string
    r = {}
    if '1'==fields[1]: r['TransA']=True
    else: r['TransA']=False
    if '1'==fields[2]: r['TransB']=True
    else: r['TransB']=False
    if '1'==fields[3]: r['Restart']=True
    else: r['Restart']=False
    if '1'==fields[4]: r['Reset']=True
    else: r['Reset']=False
    r['nmeaCmd']='CAB'
    r['nmeaPrefix']=fields[0][1:3]
    return r
    
def verQuery(prefix='xx',appendEOL=True):
    '''
    Ask for the version string from a base station
    '''
    
    rStr =  '$' + prefix+'BSQ,VER'
    rStr += '*' + checksumStr(rStr)
    
    if appendEOL: rStr += EOL
    return rStr 
    

# FIX: generic encode query function goes here


txrxLUT={
    0: 'tx a and b, rx on a and b'
    ,1: 'tx a, rx a and b'
    ,2: 'tx b, rx a and b'
    ,3: 'no tx, rx a and b'
    ,4: 'no tx, rx a'
    ,5: 'no tx, no rx'
}
'''
Transmit and Received modes.  See Page 88 61993-2 and XXXX???
'''

acaInfoSrcLUT = {
    'A':'ITU-R M.1371 message 22: addressed message'
    ,'B':'ITU-R M.1371 message 22: broadcast message'
    ,'C':'IEC 61162-1 AIS Channel Assignment setence'
    ,'D':'DSC Channel 70 telecommand'
    ,'M':'Operator manual input'
}


def acaDecode(msg,validate=True):
    '''
    Decode AIS Regional Channel Assignment Message.
    See 61993-2 Page 87.


    This is an example of an unconfigured base station, plus there is a USCG timestamp at the end.

    >>> acaDecode('$AIACA,0,,,,,,,,,5,2087,0,2088,0,0,0,I,1,000000*15,1172786646.1')
    {'inuse': '1', 'north': None, 'txrxMode': '0', 'power': '0', 'nmeaPrefix': 'AI', 'timeinuse': '000000', 'seqnum': '0', 'chanBbandwidth': '0', 'nmeaCmd': 'ACA', 'chanAbandwidth': '0', 'west': None, 'transitionSize': '5', 'infosrc': 'I', 'east': None, 'chanA': '2087', 'south': None, 'chanB': '2088'}

    @todo: get a complete example to decode as a doctest
    '''
    if validate and not isChecksumValid(msg,verbose=True):
        print 'FIX: this should be an exception in acaDecode.  Bad checksum'
        return False
    #assert msg[0]=='$'
    fields = msg.split(',')
    r = {}
    r['nmeaCmd']='ACA'
    r['nmeaPrefix']=fields[0][1:3]
    r['seqnum'] = fields[1]
    if len(fields[2])>0:
        lat = float(fields[2])  # FIX: what format is this??
        assert len(fields[3])==1
        if fields[3]=='S': lat= -1 * lat
    else: lat = None # empty string
    r['north']=lat
    del lat

    if len(fields[4])>0:
        lon = float(fields[4])  # FIX: what format is this??
        assert len(fields[5])==1
        if fields[5]=='W': lat= -1 * lat
    else: lon = None # empty string
    r['east']=lon
    del lon

    if len(fields[6])>0:
        lat = float(fields[6])  # FIX: what format is this??
        assert len(fields[7])==1
        if fields[7]=='S': lat= -1 * lat
    else: lat = None # empty string
    r['south']=lat
    del lat

    if len(fields[8])>0:
        lon = float(fields[8])  # FIX: what format is this??
        assert len(fields[9])==1
        if fields[9]=='W': lat= -1 * lat
    else: lon = None # empty string
    r['west']=lon
    del lon

    r['transitionSize'] = fields[10]  # Transition zone size in nm
    r['chanA']          = fields[11]  # Should probably be 2087
    r['chanAbandwidth'] = fields[12] # 12.5 or 25 kHz
    r['chanB']          = fields[13]  # Should probably be 2087
    r['chanBbandwidth'] = fields[14] # 12.5 or 25 kHz
    r['txrxMode']       = fields[15] # See lookup table txrxLUT
    r['power']          = fields[16] # 0 low, 1 hight
    r['infosrc']        = fields[17] # See acaInfoSrcLUT, can also be null
    r['inuse']          = fields[18] # 0 is not in use, 1 in-use.  Can also be null
    r['timeinuse']      = fields[19].split('*')[0] # hhmmss.ss
    return r


def cbmDecode(msg,validate=True):
    '''
    Decode Configure Base Station Message Broadcast Reporting Rates message.


    >>> cbmDecode('$AICBM,61,76,35,2,60,999,100,999,52,999,1,60,999,100,999*55,1172787005.46')
    {'msg17chanAnumslots': '1', 'nmeaPrefix': 'AI', 'msg4slot': '61', 'msg17chanAslotinterval': '999', 'nmeaCmd': 'CBM', 'msg20chanAslotinterval': '999', 'msg20chanAstartslot': '60', 'msg17chanAstartslot': '52', 'msg22chanAslotinterval': '999', 'msg22chanAstartslot': '100'}

    @see: 62320-1/CDV 80/427/CDV page 78, A.1.8
    '''

    if validate and not isChecksumValid(msg,verbose=True):
        print 'FIX: this should be an exception in acaDecode.  Bad checksum'
        return False
    #assert msg[0]=='$'
    fields = msg.split(',')
    r = {}
    r['nmeaCmd']=fields[0][3:] # CBM
    r['nmeaPrefix']=fields[0][1:3]
    r['msg4slot'] = fields[1]
    i = 2
    r['msg17chanAstartslot'] = fields[i]; i+=1
    r['msg17chanAslotinterval'] = fields[i]; i+=1
    r['msg17chanAnumslots'] = fields[i]; i+=1
    r['msg20chanAstartslot'] = fields[i]; i+=1
    r['msg20chanAslotinterval'] = fields[i]; i+=1
    r['msg22chanAstartslot'] = fields[i]; i+=1
    r['msg22chanAslotinterval'] = fields[i]; i+=1

    r['msg17chanAstartslot'] = fields[i]; i+=1
    r['msg17chanAslotinterval'] = fields[i]; i+=1
    r['msg17chanAnumslots'] = fields[i]; i+=1
    r['msg20chanAstartslot'] = fields[i]; i+=1
    r['msg20chanAslotinterval'] = fields[i]; i+=1
    r['msg22chanAstartslot'] = fields[i]; i+=1
    r['msg22chanAslotinterval'] = fields[i].split('*')[0]; i+=1  # 15
    return r


ownershipLUT = {
    'L':'local'
    ,'R':'remote'
    ,'C':'clear reservation'
}

def dlmDecode(msg, validate=True):
    '''
    Decode Data Link Management slot allocation for Base Station nmea message

    
    >>> dlmDecode ('$AIDLM,0,A,L,0,2,7,540,L,4,1,7,250,L,2511,1,7,0,,,,,*40,1172787005.5')
    {'nmeaPrefix': 'AI', 'timeout3': '7', 'timeout2': '7', 'timeout1': '7', 'timeout4': '', 'startslot2': '4', 'startslot3': '2511', 'incr4': '*40', 'incr3': '0', 'incr2': '250', 'incr1': '540', 'aisChannel': 'A', 'seqNum': '0', 'startslot1': '0', 'startslot4': '', 'nmeaCmd': 'DLM', 'ownership4': '', 'ownership3': 'L', 'ownership2': 'L', 'ownership1': 'L', 'numslots4': '', 'numslots1': '2', 'numslots2': '1', 'numslots3': '1'}

    @see: 62320-1/CDV 80/427/CDV page 79, A.1.9

    '''

    if validate and not isChecksumValid(msg,verbose=True):
        print 'FIX: this should be an exception in acaDecode.  Bad checksum'
        return False
    #assert msg[0]=='$'
    fields = msg.split(',')
    r = {}
    r['nmeaCmd']=fields[0][3:] # CBM
    r['nmeaPrefix']=fields[0][1:3]
    i = 1

    r['seqNum'] = fields[i]; i+=1
    r['aisChannel'] = fields[i]; i+=1

    # reservations
    r['ownership1'] = fields[i]; i+=1   # See ownership
    r['startslot1'] = fields[i]; i+=1
    r['numslots1']  = fields[i]; i+=1
    r['timeout1']   = fields[i]; i+=1
    r['incr1']      = fields[i]; i+=1

    r['ownership2'] = fields[i]; i+=1
    r['startslot2'] = fields[i]; i+=1
    r['numslots2']  = fields[i]; i+=1
    r['timeout2']   = fields[i]; i+=1
    r['incr2']      = fields[i]; i+=1

    r['ownership3'] = fields[i]; i+=1
    r['startslot3'] = fields[i]; i+=1
    r['numslots3']  = fields[i]; i+=1
    r['timeout3']   = fields[i]; i+=1
    r['incr3']      = fields[i]; i+=1

    r['ownership4'] = fields[i]; i+=1
    r['startslot4'] = fields[i]; i+=1
    r['numslots4']  = fields[i]; i+=1
    r['timeout4']   = fields[i]; i+=1
    r['incr4']      = fields[i]; i+=1

    return r

def bbmEncode(totSent, sentNum, seqId, aisChan, msgId, data, numFillBits 
              ,prefix='xx',appendEOL=True
              ,validate=True
              ):
    '''Encode a binary broadcast message.

    I have no idea what this message says...

    !AIVDM,1,1,,A,85NqMF1Kf=Vsdt`l;0bnfFjd<uQeT2p<vmIRTB=mM5mtIT;sUL2t,0*54,rs003669982,1172918061

    >>> bbmEncode(1,1,0,3,8,'Fs[Ifs?:=2h:ec]dc3?HKI0f3?eFHa4[MGAMO6I2vqG0g',4)
    '!xxBBM,1,1,0,3,8,Fs[Ifs?:=2h:ec]dc3?HKI0f3?eFHa4[MGAMO6I2vqG0g,4*32'

    @todo: put in some doc tests with know messages and what would be received as the VDM message(s)
    @see: IEC-PAS 61162-100 80/330/PAS, Page 19
    @param totSent: Total number of sentences needed for the message (1-9)
    @type totSent: int
    @param sentNum: Which sentence is this in the series (1-9)
    @type sentNum: int
    @param seqId: need to increment this for each message??!?!?  (0-9)  Linked to ABK
    @type seqId: int
    @param aisChan: AIS channel to use to send the message
      0. No channel preference
      1. AIS Channel A
      2. AIS Channel B
      3. Broadcast on both A and B
    @type aisChan: str
    @param msgId: AIS message 8 (binary broadcast message) or 14 (safety related broadcast)
    @type msgId: int
    @param data: Content of the binary data.  First sentence must be 58 characters or less.  
    The rest can be up to 60 characters.
    @param numFillBits: Number of bits of padding in the last character of the data (0-5)
    @type numFillBits: int
    @return: nmea string
    @rtype: str
    '''
    if validate:
        # obsesive error checking follows
        tot = int(totSent)
        assert (0<tot and tot<=9)
        num = int(sentNum)
        assert (0<num and num<=9)
        assert (num<=tot)
        seq = int(seqId)
        assert (0<=seq and seq <=9)
        assert (int(aisChan) in range(0,5))
        assert (int(msgId) in (8,14))
        if num==1: assert(len(data)<=58)
        else: assert(len(data)<=60)
        # FIX: add validation of the characters in the data string
        assert(int(numFillBits) in range(0,6))
    r = ','.join(('!'+prefix+'BBM',str(totSent),str(sentNum),str(seqId),str(aisChan),str(msgId),data,str(numFillBits)))

    r += '*'+checksumStr(r)
    if validate: assert(len(r)) <= 81  # Max nmea string length
    return r


def bbmDecode(msg,validate=True):
    '''
    Decode a binary broadcast message NMEA string

    >>> bbmDecode('!xxBBM,1,1,0,3,8,Fs[Ifs?:=2h:ec]dc3?HKI0f3?eFHa4[MGAMO6I2vqG0g,4*32')
    {'numFillBits': '4', 'nmeaPrefix': 'xx', 'msgId': '8', 'aisChan': '3', 'data': 'Fs[Ifs?:=2h:ec]dc3?HKI0f3?eFHa4[MGAMO6I2vqG0g', 'seqId': '0', 'nmeaCmd': 'BBM', 'sentNum': '1', 'totSent': '1'}

    @todo: make the doctest stable
    @todo: doctests with known messages
    @param msg: NMEA string of a CAB message
    @type msg: str
    @param validate: Set to False to turn off validation for speed.
    @type validate: bool
    @return: lookup table of key/values
    @rtype: dict
    @see: IEC-PAS 61162-100 80/330/PAS, Page 19
    '''
    if validate and not isChecksumValid(msg,verbose=True):
        print 'FIX: this should be an exception in cabDecode.  Bad checksum'
        return False
    fields=msg.split(',')
    if validate and len(fields) < 8: # Allow for USCG station and timestamp
        print 'FIX: this should be an exception in cabDecode.  wrong number of fields', len(fields)
        print '  ', msg
        return False

    r = {}

    fields = msg.split(',')
    r = {}
    r['nmeaCmd']=fields[0][3:] # CBM  # 0
    r['nmeaPrefix']=fields[0][1:3]
    i = 1
    # FIX: convert from strings or not??
    r['totSent'] = fields[1]; i+= 1 
    r['sentNum'] = fields[2]; i+= 1
    r['seqId'] = fields[3]; i+= 1
    r['aisChan'] = fields[4]; i+= 1
    r['msgId'] = fields[5]; i+= 1
    r['data'] = fields[6]; i+= 1
    r['numFillBits'] = fields[7].split('*')[0]; i+= 1

    return r

######################################################################
if __name__=='__main__':
    from optparse import OptionParser
    myparser = OptionParser(usage="%prog [options]",
			    version="%prog "+__version__)
    myparser.add_option('--test','--doc-test',dest='doctest',default=False,action='store_true',
                        help='run the documentation tests')
    myparser.add_option('-v','--verbose',dest='verbose',default=False,action='store_true',
                        help='run the tests run in verbose mode')
    (options,args) = myparser.parse_args()

    success=True

    if options.doctest:
	import os; print os.path.basename(sys.argv[0]), 'doctests ...',
	argvOrig = sys.argv
	sys.argv= [sys.argv[0]]
	if options.verbose: sys.argv.append('-v')
	import doctest
	numfail,numtests=doctest.testmod()
	if numfail==0: print 'ok'
	else: 
	    print 'FAILED'
	    success=False
	sys.argv = argvOrig # Restore the original args
	del argvOrig # hide from epydoc

    if not success:
	sys.exit('Something Failed')

    del success # Hide success from epydoc
